from .nadcon5_input import RegionData, ControlData, InData, ExclusionData
from .nadcon5_output import VectorData, PointData
from .nadcon5_types import MetaMixin, GMTMetaMixin
from .services import region_bounds
import logging
import itertools


class Conversion(MetaMixin, GMTMetaMixin):
    """ A Conversion agregates all parts of a NADCON5 Datum Conversion and serves
    as the primary user interface into the dataset

    A Conversion is created based on region, source,  target datum, and gridspacing
    the same parameters that go into the data build pipeline

    Conversions maintain a large set of data in memory (when loaded) and have accessors for data

    Creating a conversion is simple

        c = Conversion('conus', 'ussd', 'nad27')

    The output data of a conversion can be accessed directly by output prefix

       v1 = c.output_data['vmacdlat']
       v2 = c.output_data['vmacdlon']

    Output data is indexed by PID, to extract all PID's with lat and lon conversions in this data set

       shared_pids = [ i for i in v1 if i in v2 ] 

    
    All point data for a single point can be examined directly, including its source data

       point = v[shared_pids[0]]
       point.source 


    """

    class ConversionInput(object):

        @property
        def region_data(self):
            return self._region_data
        
        @property
        def grid_bound(self):
            return self._grid_bound

        @property
        def control_data(self):
            return self._control_data

        @property
        def input_data_set(self):
            return self._input_data_set

        @property
        def input_pid_set(self):
            return self._input_pid_set

        @property
        def input_point_set(self):
            return self._input_point_set

        @property
        def exclusion_data(self):
            return self._exclusion_data

        @property
        def exclusion_pid_set(self):
            return self._exclusion_pid_set

        @property
        def exclusion_point_set(self):
            return self._exclusion_point_set

        @property
        def pruned_point_set(self):
            return self._pruned_points

        
        def __init__(self, region, old_datum, new_datum, **kwargs):
            self._region_data = RegionData()
            self._grid_bound = self._region_data[region]

            self._control_data = ControlData(region, old_datum, new_datum)

            self._input_pid_set = set()
            self._input_data_set = set()
            self._input_point_set = set()
            for in_point in self._control_data:
                i = InData(in_point.in_file)
                [self._input_pid_set.add(_) for _ in i.indices]
                [self._input_point_set.add(_) for _ in i.points]
                self._input_data_set.add(i)

            self._exclusion_data = ExclusionData()
            self._exclusion_pid_set = set()
            self._exclusion_point_set = set()
            for exclusion_point in self._exclusion_data:
                if (exclusion_point.olddtm == old_datum) and (exclusion_point.newdtm == new_datum) and (exclusion_point.region == region):
                    self._exclusion_pid_set.add(exclusion_point.pid)
                    self._exclusion_point_set.add(exclusion_point)

            self._pruned_points = {_p for _p in self._input_point_set if not _p.pid in self._exclusion_pid_set}

    class ConversionOutput(object):
        def __init__(self, region, old_datum, new_datum, grid_spacing, load_all = False, **kwargs):
            self._output_data = {}
            vdir = ['lat', 'lon', 'eht','hor']
            vclass = ['a','t','d','r']
            vout = ['cd','dd','gi']
            vunit = ['m', 's']
            surface = [True, False]
            

            c_pars = itertools.product(vdir, vclass,vout)
            v_pars = itertools.product(vdir, vclass, vout,vunit,surface)

            local_kwargs = {}
            if 'out_fdir' in kwargs:
                local_kwargs['fdir'] = kwargs['out_fdir']

            print ("Loading Coverage Files")
            for c in c_pars:
                print ("Trying to Load Coverage File args - %s"%str(c))
                try:
                    cov = PointData(region,old_datum, new_datum, grid_spacing, *c, **local_kwargs)
                    print ("Finished Loading Coverage File - %s"%cov.shorthand)
                except (ValueError, FileNotFoundError) as e:
                    print ("Could not find Coverage File")
                    print (e)
                    continue
                self._output_data[cov.shorthand] = cov

            print ("Loading Vector Files")
            for v in v_pars:
                print ("Trying to Load Vector File args - %s"%str(v))
                try:
                    vec = VectorData(region, old_datum, new_datum, grid_spacing, *v, **local_kwargs)
                    print ("Finished Loading Vector File - %s"%vec.shorthand)
                except (ValueError, FileNotFoundError) as e:
                    print ("Could not find Vector File")
                    print (e)
                    continue
                self._output_data[vec.shorthand] = vec

        
        def __contains__(self, item):
            return ( item in self.raw_data ) or ( item in self.pid_set )


        def __getitem__(self, key):
            if key in self.raw_data:
                return self.raw_data[key]
            elif key in self.pid_set:
                return { d_k:d_v[key] for d_k, d_v in self.raw_data.items() if key in d_v.indices }
            else:
                return None

        @property
        def raw_data(self):
            return self._output_data

        @property
        def pid_set(self):
            return self._pid_set

        
            
            

            


    @property
    def input_data(self):
        return self._input_data

    @property
    def output_data(self):
        return self._output_data
            
    def __init__(self, region, old_datum, new_datum, grid_spacing = '900',  **kwargs):
        self._input_data = self.ConversionInput(region, old_datum, new_datum, **kwargs)
        self._output_data = self.ConversionOutput(region, old_datum, new_datum, grid_spacing, **kwargs)
        self._meta={'region':region,
                    'old_datum':old_datum,
                    'new_datum':new_datum,
                    'grid_spacing':grid_spacing,
                    **kwargs}


        self._gmt_meta={
            'region':region_bounds(region)
        }

        
            
        if 'gmt' in kwargs:
            self._gmt_meta.update(kwargs['gmt'])



from .nadcon5_files import VectorFileParser, FileBackedMetaBase, CoverageFileParser
from .services import region_bounds
from .nadcon5_types import DataPoint, DataContainerMixin, MetaMixin, GMTMixin, GMTPointMixin
import numpy as np
import pkg_resources




class VectorData(DataContainerMixin, MetaMixin, GMTMixin, metaclass = FileBackedMetaBase, Parser = VectorFileParser):
    """@ingroup nc5data

    """
        

    class VectorDataPoint(DataPoint, GMTPointMixin):
        def __init__(self, lon, lat, az, vector, dlatsec, dlatm, pid, **kwargs):
            super().__init__(lon=lon, lat=lat, az=az, vector=vector, dlatsec=dlatsec, dlatm=dlatm, pid=pid, **kwargs)
        def __mk_gmt_data__(self):
            return (self.lon, self.lat, self.az, self.vector)
            
    def __init_indexed_data__(self):
        self._indexed_data = { data[-1]:self.VectorDataPoint(*data, **self.meta) for data in self.data}


    def __mk_plot_args__(self):
        return {'data':np.array(self.data)[:,:4].astype(float),  'region':region_bounds(self.region), style:'V'}



class PointData(DataContainerMixin, MetaMixin, GMTMixin, metaclass=FileBackedMetaBase, Parser = CoverageFileParser):

    class PointDataPoint(DataPoint, GMTPointMixin):
        def __init__(self, lon, lat, _empty_, pid, **kwargs):
            super().__init__(lon=lon, lat=lat,  pid=pid, **kwargs)

        def __mk_gmt_data__(self):
            return (self.lon, self.lat)
            

    def __init_indexed_data__(self):
        self._indexed_data = { data[-1]:self.PointDataPoint(*data, **self.meta) for data in self.data}


    def __mk_plot_args__(self):
        return {'data':np.array(self.data)[:,:3].astype(float), 'region':region_bounds(self.region), style:'P'}
        


class GRDData(object):
    pass

class BData(object):
    pass

        
class XYZData(object):
    pass



__path__ = __import__('pkgutil').extend_path(__path__, __name__)

# This is required to fill-in the nc5ng namespace
# The source in this folder only creates the core utilities
# Other projects provide additional subpakcages in nc5ng.*




from .nadcon5_input import RegionData



s = RegionData()

def region_bounds(region):
    """ Compute Region Bounds from singleton data"""
    if region in s:
        return s[region].bounds
    else:
        return None


def regions():
    return list(s.indices)


    



from .conversion import Conversion
from .nadcon5_input import RegionData, ControlData, InData, ExclusionData
from .nadcon5_output import VectorData, PointData, GRDData, BData,XYZData
from .nadcon5_types import DataPoint
from .nadcon5_gmt import GMTPlotter
from .services import region_bounds






__all__ = ['Conversion', 'RegionData', 'ControlData', 'InData', 'ExclusionData',
           'VectorData', 'PointData', 'GRDData', 'BData', 'XYZData', 'DataPoint', 'GMTPlotter', 'region_bounds']



""" \file nadcon5_files_ng.py
"""
from os.path import basename, exists, join, isdir, isfile, isabs
from os import listdir
import pkg_resources

from .utils import dmstodec, output_filename

import fortranformat as ff
import logging


NADCON5_FILE_RESOURCES = {
    'in_dir': pkg_resources.resource_filename('nc5ng.nc5data', 'data/InFiles'),
    'grid_file': pkg_resources.resource_filename('nc5ng.nc5data', 'data/Data/grid.parameters'),
    'control_dir': pkg_resources.resource_filename('nc5ng.nc5data', 'data/Control'),
    'workedits_file': pkg_resources.resource_filename('nc5ng.nc5data', 'data/Work/workedits'),
}


class BaseFileParser(object):
    def __init__(self, parser=None, fdir=None, ffile=None):
        self.parser = parser
        self.fdir = fdir
        self.ffile = ffile
        
        

    @property
    def parser(self):
        return self._parser
    @parser.setter
    def parser(self, value):
        self._parser = value

    @property
    def fdir(self):
        return getattr(self, '_fdir', None)

    @fdir.setter
    def fdir(self, value):
        self._fdir = value
        if (value is not None) and not(isabs(value)):
            logging.warning("%s is not an absolute path"%str(value))

    @property
    def ffile(self):
        return self._ffile

    @ffile.setter
    def ffile(self, value):
        self._ffile = value
        

    def __call__(self, it):
        if self.parser and hasattr(self.parser, '__call__'):
            return [self.parser(_line) for _line in it]
        elif self.parser and hasattr(self.parser, 'read'):
            return [self.parser.read(_line) for _line in it]
        else:
            return None

    def __fromfile__(self, f):
        return {'meta': {}, 'data':[ _ for _ in self(f) if _ is not None] }

    def fromfile(self, ffile=None, process=None, fdir=None):
        if ffile is None and self.ffile is not None:
            ffile = self.ffile

        if fdir is None and self.fdir is not None:
            fdir = self.fdir
        
        if (fdir is not None) and (ffile is not None) and  not(isabs(ffile)):
            ffile = join(fdir, ffile)

        with open(ffile,'r') as f:
            res = self.__fromfile__(f)
            res['meta']['source'] = ffile
            return res
        
        return None

class FortranFormatFileParser(BaseFileParser):
    def __init__(self, fformat = None, ffilter = None):
        self._parser = None
        self.fformat = fformat
        self.ffilter = ffilter
        
    @property
    def fformat(self):
        return self._fformat
    
    @fformat.setter
    def fformat(self, value):
        self._fformat = value
        if self._fformat:
            self._parser = ff.FortranRecordReader(self._fformat)

    @property
    def ffilter(self):
        return self._ffilter

    @ffilter.setter
    def ffilter(self, value):
        if value is None:
            self._ffilter = lambda x: x
        else:
            self._ffilter = value            
    
    def __call__(self, it):
        if self._parser:
            return [ self._ffilter( self._parser.read(_line) ) for _line in it ]
    
        

class IndexedFortranFormatFileParser(FortranFormatFileParser):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._running_index=0
        self._index=0
        self.indexed_format = {}
        
        for entry in args:
            try:
                fformat,ffilter = entry
            except TypeException:
                fformat = args.pop(0)
                if args: ffilter = args.pop(0)
                else: ffilter=None
            self._register_format(fformat, ffilter)
        for index, entry in kwargs.items():
            try:
                fformat, ffilter = entry
            except TypeException:
                fformat = entry
                ffilter = None
            self._register_format(fformat, ffilter, index)
        self.index=0

    def _register_format(self, fformat, ffilter=None, index = None, overwrite=False):
        if index is None:
            index = self._running_index
            self.fformat = fformat
            self._running_index = self._running_index + 1

        if (index not in self.indexed_format or overwrite):
            self.indexed_format[index]=(fformat,ffilter)
            if (index == self.index):
                self.index = index # reload filters
            return index
        else:
            return None

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, index=None):
        if index in self.indexed_format:
            fformat, ffilter = self.indexed_format[index]
            self.fformat = fformat
            self.ffilter = ffilter
            self._index = index
        else:
            pass
        
    def __getitem__(self,index):
        if index in self.indexed_format:
            return FortranFormatFileParser(*self.indexed_format[index])
        else:
            return None

    def __contains__(self, value):
        return index in self.indexed_format

    def __call__(self, it, index=None):        
        if index is None or index == self.index:
            return super().__call__(it)
        elif index in self:
            return p(it)
        else:
            return None
                        
    



class InFileParser(IndexedFortranFormatFileParser):
    IN_FILE_HEADER_FORMAT = '27x,a15,26x,a15'
    IN_FILE_DATA_FORMAT = 'a6,1x,a2,5x,a13,1x,a14,1x,a9,3x,a13,1x,a14,1x,a9'

    @staticmethod
    def _line_filter(line):
        pid, subr, oldlat, oldlon, oldht, newlat, newlon, newht = line
        oldlat = None if oldlat.strip() == "N/A" else dmstodec(oldlat)
        oldlon = None if oldlon.strip() == "N/A" else dmstodec(oldlon)
        oldht = None if oldht.strip() == "N/A" else float(oldht)
        newlat = None if newlat.strip() == "N/A" else dmstodec(newlat)
        newlon = None if newlon.strip() == "N/A" else dmstodec(newlon)
        newht = None if newht.strip()  == "N/A" else float(newht)

        return [pid,subr,oldlat,oldlon,oldht,newlat,newlon,newht]

    @staticmethod
    def _header_filter(line):
        return [ _.strip() for _ in line]
    
    def __init__(self, fdir=NADCON5_FILE_RESOURCES['in_dir'], ffile = None):
        self.fdir = fdir
        self.ffile = ffile
        super().__init__(data=(self.IN_FILE_DATA_FORMAT, InFileParser._line_filter), header=(self.IN_FILE_HEADER_FORMAT, InFileParser._header_filter))
        
    def __fromfile__(self, f):        
        header = self['header']([next(f)])[0]
        meta={'DATUM1':header[0], 'DATUM2':header[1], 'header':header}
        return {'meta':meta, 'data':self['data'](f)}
    
    def fromfile(self, ffile=None, old_datum=None, new_datum=None, subregion=None):
        if ffile is not None:
            res = super().fromfile(ffile)
        elif old_datum and new_datum:
            old_datum = old_datum.lower()
            new_datum = new_datum.lower()
            subregion = subregion.lower() if subregion is not None else None
            if subregion:
                res = super().fromfile( "NADCON5.%s.%s.%s.in"%(old_datum, new_datum, subregion))
            else:
                res = super().fromfile( "NADCON5.%s.%s.in"%(old_datum, new_datum ))            
        else:
            raise TypeError("fromfile Incorrect number of arguments")

        res['meta'].update({'subregion':subregion, 'old_datum':old_datum, 'new_datum':new_datum})
        return res


    

        
class ControlFileParser(BaseFileParser):
    @staticmethod
    def _parse_line(line):
        if '.in' in line:
            return line.strip()
        else:
            return None

    def __init__(self, control_dir = NADCON5_FILE_RESOURCES['control_dir']):
        super().__init__(ControlFileParser._parse_line, fdir = control_dir)

    def __fromfile__(self, f):
        header_lines = ['HEADER','REGION', 'DATUM1', 'DATUM2', 'REJMET', 'NFILES']
        
        data = []
        meta = {}

        for l in f:
            if '.in' not in l:
                
                h,v = (l[:6], l[7:].strip(),)
                if h in header_lines:
                    meta[h.lower()] = v.lower()
                else:
                    logging.warning('Encountered Control File Header Line that is not Known')
                
            else:
                data.append(ControlFileParser._parse_line(l))

        return {'meta':meta, 'data':data}
                
    def fromfile(self, *args):

        argc = len(args)
        if argc == 1:
            res = super().fromfile(args.pop())
        elif argc == 3:
            region, old_datum, new_datum = args
            res = super().fromfile("control.%s.%s.%s"%(old_datum.lower(), new_datum.lower(),region.lower() ))
        else:
            raise TypeError("fromfile Incorrect number of arguments")

        res['meta'].update({'region':region, 'old_datum':old_datum, 'new_datum':new_datum})
        return res



class GridParamFileParser(BaseFileParser):
    _VALID_REGIONS = (
        'conus',
        'alaska',
        'hawaii',
        'prvi',
        'as',
        'guamcnmi',
        'stpaul',
        'stgeorge',
        'stlawrence',
        'stmatthew',
    )
    """ Hardcoded Valid Regions """

    def gen_parse_line(self):
        _regions = self._VALID_REGIONS
        def parse_line(line):
            region, *data = line.split()
            
            if region.lower() in _regions:
                return [region,]+ [ _ for _ in map (lambda x: float(x.strip()), data)]
            else:
                return None

        return parse_line
        
    
    
    def __init__(self, grid_file = NADCON5_FILE_RESOURCES['grid_file'], *regions):
        if regions:
            self._VALID_REGIONS = self._VALID_REGIONS + regions
        super().__init__(self.gen_parse_line(), ffile=grid_file)
    
    def valid_region(self, region):
        return region in self._VALID_REGIONS

class WorkEditsFileParser(BaseFileParser):
    @staticmethod
    def _parse_line(line):
        """
        01- 10  : olddtm : lower case, left justified 
        11  : "|"    : vertcal spacer just for ease of reading
        12- 21  : newdtm : lower case, left justified 
        22  : "|"    : vertcal spacer just for ease of reading
        23- 32  : region : lower case, left justified (conus, alaska, hawaii, prvi, guamcnmi, as)
        33  : "|"    : vertcal spacer just for ease of reading
        34- 39  : PID    : upper case, left justified 
        40  : "|"    : vertcal spacer just for ease of reading
        41- 43  : rejects: Three digits (0's or 1's only) to reject lat, lon, eht, in that order.  '1' = reject, '0' = keep
        44  : "|"    : vertcal spacer just for ease of reading
        45-200  : reason : Upper/lower case, giving first your name then reason for the line to exist
        """
        
        if (len(line) < 44) or not(line[10] == line[21] == line[32] == line[39] == line[43] == '|'):
            return None

        
        
        olddtm = line[:10].strip()
        newdtm = line[11:21].strip()
        region = line[22:32].strip()
        pid = line [33:39].strip()
        rejlat = line[40] == '1'
        rejlon = line[41] == '1'
        rejeht = line[42] == '1'
        reason = line[44:]

        return [olddtm, newdtm, region, pid, rejlat, rejlon, rejeht, reason]

    def __init__(self, ffile = NADCON5_FILE_RESOURCES['workedits_file']):
        super().__init__(self.__class__._parse_line, ffile=ffile)


class CoverageFileParser(FortranFormatFileParser):
    COVERAGE_FILE_FORMAT = "f16.10,1x,f15.10,1x,f6.2,1x,a6"
    def __init__(self, **kwargs):
        super().__init__(fformat=self.COVERAGE_FILE_FORMAT, **kwargs)


    def fromfile(self,
                 region='conus',
                 old_datum='ussd',
                 new_datum='nad27',
                 grid_spacing='900',
                 vdir='lon',
                 vclass='a',
                 vout='cd',
                 **kwargs):


        if 'ffile' in kwargs:
            return super().fromfile(**kwargs)


        meta_dict = {'old_datum':old_datum,
                     'new_datum':new_datum,
                     'region':region,
                     'grid_spacing':grid_spacing,
                     'vclass':vclass,
                     'vout':vout,
                     'vdir':vdir}

        ffile = output_filename(output_type = 'c', **meta_dict)

        meta_dict['basename'] = ffile.split('.')[0]
        
        res = super().fromfile(ffile=ffile, **kwargs)
        
        res['meta'].update(meta_dict)

        return res

        

class VectorFileParser(FortranFormatFileParser):
    VECTOR_FILE_FORMAT = "f16.10,1x,f15.10,1x,f6.2,1x,f12.2,1x,f9.5,1x,f9.3,1x,a6"
    
    def __init__(self, **kwargs):
        super().__init__(fformat=self.VECTOR_FILE_FORMAT, **kwargs)

    def fromfile(self,
                 region='conus',
                 old_datum='ussd',
                 new_datum='nad27',
                 grid_spacing='900',
                 vdir='lon',
                 vclass='a',
                 vout='cd',
                 vunit='m',
                 surface=False,
                 **kwargs):

        if 'ffile' in kwargs:
            return super().fromfile(**kwargs)
            
    
        meta_dict = {'old_datum':old_datum,
                     'new_datum':new_datum,
                     'region':region,
                     'grid_spacing':grid_spacing,
                     'vunit':vunit,
                     'vclass':vclass,
                     'vout':vout,
                     'vdir':vdir}


        ffile = output_filename(output_type = 'v', surface=surface, **meta_dict)

        meta_dict['basename'] = ffile.split('.')[0]

        res = super().fromfile(ffile=ffile, **kwargs)

        res['meta'].update(meta_dict)
        return res

        

class FileBackedMetaBase(type):
    @classmethod
    def __prepare__(metacls, name, bases, **kargs):
        """ Prepare the new class, here for completeness
        """
        logging.debug("Preparing Class %s"%name)
        return super().__prepare__(name, bases, **kargs)

    def __new__(metacls, name, bases, namespace, **kargs):
        logging.debug("Creating Class %s"%name)
        return super().__new__(metacls, name, bases, namespace)

    @property
    def parser(self):
        return self._parser
    

    @parser.setter
    def parser(self, parser):
        logging.debug("Setting parser instance %s"%parser)
        if not(issubclass(parser.__class__,BaseFileParser)):
            raise TypeError("parser %s is not a valid BaseFileParser"%str(parser))

        self._parser = parser

    @property
    def instances(self):
        return self._instances

    
    def __init__(cls, name, bases, namespace, Parser=BaseFileParser, **kwargs):
        """ Initialize a new FileBacked Class

        This is a slot method for class creation, __init__ is called when class is defined (load time)

        \param cls - reference to new type, similiar to @classmethod
        \param name - new class name
        \param bases - base classes
        \param namespace - new class attributes
        \param Parser - BaseFileParser underlying this type
        \param **kwargs - keywords passed to Parser initialization
        
        """        
        logging.debug("Creating File Backed Class %s"%name)

        if Parser:
            logging.debug("Creating Parser Instance of %s"%Parser)
            cls.parser = Parser(**kwargs)
            
        cls._Parser = Parser
        cls._instances = dict()
        super().__init__(name, bases, namespace)
            

class SingletonFileBackedMeta(FileBackedMetaBase):
    
    def __init__(cls, name, bases, namespace, Parser=BaseFileParser, **kwargs):
        """
        """        
        cls._instance = None
        super().__init__(name, bases, namespace, Parser, **kwargs)

    def __call__(cls, *args, overwrite=False, **kwargs):
        if cls._instance is None or overwrite:
            logging.debug("Creating new Singleton File Backed Meta Instance %s"%cls.__name__ )
            inst = super().__call__(*args, **kwargs)
            cls._instance = inst
            return inst
        else:
            return cls._instance
        




from .nadcon5_files import ControlFileParser, InFileParser, FileBackedMetaBase, WorkEditsFileParser, SingletonFileBackedMeta, GridParamFileParser
from .nadcon5_types import DataPoint, DataContainerMixin, MetaMixin, GMTMixin, GMTPointMixin
import pkg_resources


class RegionData(DataContainerMixin, MetaMixin, metaclass = SingletonFileBackedMeta, Parser = GridParamFileParser):
    """@ingroup nc5data

    """
        

    class GridBound(DataPoint):
        def __init__(self, region, n,s,w,e, **kwargs):
            super().__init__(region=region,N=n, S=s, W=w, E=e, **kwargs)
        @property
        def bounds(self):
            return (self.W, self.E, self.S, self.N)

    def __init_indexed_data__(self):
        self._indexed_data = { region:self.GridBound(region, *data, **self.meta) for region, *data in self.data}




def _extract_subregion(in_file):
    vals = in_file.split('.')
    if len(vals) == 4:
        return None
    if len(vals) == 5:
        return vals[-2]

class ControlData(DataContainerMixin, MetaMixin, metaclass=FileBackedMetaBase, Parser=ControlFileParser):
    class ControlDataPoint(DataPoint):
        def __init__(self, in_file, **kwargs):
            
            super().__init__( in_file = in_file, subregion=_extract_subregion(in_file), **kwargs)
            
        


    def __init_indexed_data__(self):
        self._indexed_data = [ ControlData.ControlDataPoint(in_file, **self.meta)  for in_file in self.data ]
        self._index_tag = 'subregion'
        self._indexed_data = { point.subregion : point for point in self._indexed_data }
        

    @property
    def rejmet(self):
        return int(self._data['meta']['rejmet'])

    @property
    def n_infiles(self):
        return int(self._data['meta']['nfiles'])
        
    @property
    def infiles(self):
        return self._data['data']

            

    def __contains__(self, point):
        
        if not(super().__contains__(point)) and 'region' in point and point.region == self.region:
            return True
        else:
            return False




class InData(DataContainerMixin, MetaMixin,GMTMixin,  metaclass=FileBackedMetaBase, Parser=InFileParser):
            
    class InDataPoint(DataPoint):
        def __init__(self, pid, subr, oldlat, oldlon, oldht, newlat, newlon, newht, **kwargs):
            super().__init__(pid=pid, #subregion=subregion,
                             oldlat=oldlat, oldlon=oldlon, oldht=oldht, newlat=newlat, newlon=newlon, newht=newht, **kwargs)


    def __init_indexed_data__(self):
        self._index_tag = 'pid'
        self._indexed_data = { pid:InData.InDataPoint(pid, *data, **self.meta) for pid, *data in self.data}


    @property
    def source(self):
        if 'meta' in self._data and 'source' in self._data['meta']:
            return self._data['meta']['source']

    def __mk_plot_args__(self):
        return {'data':np.array(self.data)[:,2:3].astype(float), 'style':'p','region':RegionData.region_bounds(self.region)}

                 
        
            
    


class ExclusionData(DataContainerMixin, MetaMixin, metaclass=SingletonFileBackedMeta, Parser=WorkEditsFileParser):
    
    class ExclusionDataPoint(DataPoint):
        def __init__(self,  olddtm, newdtm, region,pid,rejlat, rejlon, rejeht, reason, **kwargs):
            super().__init__( olddtm=olddtm, newdtm=newdtm, region=region,pid=pid,rejlat=rejlat, rejlon=rejlon, rejeht=rejeht, reason=reason, **kwargs)


    def __init_indexed_data__(self):
        self._index_tag = 'pid'
        self._indexed_data = { (data[3], data[0], data[1]):self.ExclusionDataPoint(*data, **self.meta) for data in self.data }
    



from .nadcon5_output import VectorData, PointData
from .services import region_bounds

_pointstore = PointData.PointDataPoint.point_store
_vectorstore = VectorData.VectorDataPoint.point_store

DEFAULT_PROJECTION = 'M10.0i'
DEFAULT_REGION = [240,190,30,80]
DEFAULT_RESOLUTION='f'
DEFAULT_THRESH=None
DEFAULT_FRAME=True
DEFAULT_INSERT=None
DEFAULT_BORDER=None
DEFAULT_SCALE=None
DEFAULT_DIR_ROSE=None
DEFAULT_MAG_ROSE=None
DEFAULT_LOGO=False
DEFAULT_LAKES=None
DEFAULT_LAND=None
DEFAULT_RIVERS = None
DEFAULT_BORDERS = None
DEFAULT_WATER = None
DEFAULT_LINEAR = False
DEFAULT_CPT = None
DEFAULT_ERRORS = None
DEFAULT_COLOR = None
DEFAULT_STYLE = None

ARG_MAP = {
    'basemap':{
        'projection':'J'.
        'region':'R',
        'frame':'B',
        'insert':'D',
        'border':'F',
        'scale':'L',
        'dir_rose':'Td',
        'mag_rose':'Tm',
        'logo':'U',
    },
    'coast':{
        'area_thresh':'A',
        'lakes':'C',
        'resolution':'D',
        'land':'G',
        'rivers':'I',
        'borders':'N',
        'water':'S',
        'shorelines':'W',
    },
    'plot':{
        'cpt':'C',
        'offset':'D',
        'errors':'B',
        'color':'G',
        'style':'S',
        'pen':'W',
    },
}


class GMTOptions(dict):

    def __init__(self,
                 projection = DEFAULT_PROJECTION,
                 region = DEFAULT_REGION,
                 frame=DEFAULT_FRAME,
                 insert=DEFAULT_INSERT,
                 border=DEFAULT_BORDER,
                 scale = DEFAULT_SCALE,
                 dir_rose = DEFAULT_DIR_ROSE,
                 mag_rose = DEFAULT_MAG_ROSE,
                 logo = DEFAULT_LOGO,
                 area_thresh = DEFAULT_THRESH,
                 lakes = DEFAULT_LAKES,
                 resolution=DEFAULT_RESOLUTION,
                 land = DEFAULT_LAND,
                 rivers = DEFAULT_RIVERS,
                 borders = DEFAULT_BORDERS,
                 water = DEFAULT_WATER,
                 shorelines = DEFAULT_SHORELINES,
                 linear_lines = DEFAULT_LINEAR,
                 cpt = DEFAULT_CPT,
                 offset = DEFAULT_OFFSET,
                 errors = DEFAULT_ERRORS,
                 color = DEFAULT_COLOR,
                 style = DEFAULT_STYLE,
                 pen = DEFAULT_PEN,
                 basemap = None,
                 coast = None
                 plot = None,
                 **kwargs):
        super().__init__(**locals())
    
        self.__init_basemap__()
        self.__init_coast__()
        self.__init_plot__()


    def __init_basemap__(self):
        pass

    def __init_coast__(self):
        pass

    def __init_plot__(self):
        pass
        
                 
            

    @property
    def basemap(self):
        

    @property
    def coast(self):
        return self.get('coast',None)

    @property
    def plot(self):
        return self.get('plot',None)


    

    
PLOT_OPTS = {
    'default': {
        'projection':"M10.0i",
        'frame':True,
        'resolution':'fine',
        'water':"lightblue",
        'borders':["1", "2"],
        'area_thresh':1200,
    },

    'vector':{
        'style':"V0.0001i/0.02i/0.02i"
    },
    'point':{
        'style':"P0.01i"
    },

    'red':{
        'color':'red'
    },
    'black':{
        'color':'black'
    },

    'blue':{
        'color':'blue'
    },

}
    
    

def _mk_filter_plot_pars(conversion, names, *aliases):
    def _filter(args):
        _COAST_ARGS = ['A','C', 'D', 'G','I', 'N', 'S', 'W']
        _COAST_ARGS_ALT = ['area_thresh', 'lakes','resolution', 'land', 'rivers', 'borders',  'water', 'shorelines']
        _COAST_DICT = dict(zip(_COAST_ARGS_ALT, _COAST_ARGS))
        _filter_base_kwargs =  { x:y for x,y in kwargs.items() if x in _COAST_ARGS}
        for _arg in conversion.gmt_meta:
            if _arg in _COAST_DICT and _COAST_DICT[_arg] not in _filter_base_kwargs:
                _filter_base_kwargs[ _COAST_DICT[_arg] ] = conversion.gmt_meta[_arg]




        
        


class GMTPlotter(object):

    @property
    def figure(self):
        return getattr(self, '_figure', None)

    def __init__(self, conversion, base_plot_optons=PLOT_OPTS['default']):
        import gmt
        self._conversion = conversion
        self._figure = gmt.Figure()

    def plot(self, *args, **kwargs):
        self.figure.plot(*args, **kwargs)

    def show(self, *args, **kwargs):
        self.figure.show(*args, **kwargs)

    @staticmethod
    def plot_conversion(conversion, coverage='all', vector='all', plotter=None, **kwargs):
        if plotter is None:
            plotter = GMTPlotter()
            
        if (coverage == 'all'):
            cpoints = [_pointstore,]
        elif not(coverage):
            cpoints = []
        elif coverage[0] == 'c' and coverage in conversion.output_data:
            cpoints = [conversion.output_data[coverage],]
        else:
            cpoints = []
            for c in coverage:
                cpoints.append(conversion.output_data[c])


        if (vector == 'all'):
            vpoints = [_vectorstore,]
        elif not(vector):
            vpoints = []
        elif vector[0] == 'c' and vector in conversion.output_data:
            vpoints = [conversion.output_data[vector],]
        else:
            vpoints = []
            for c in vector:
                vpoints.append(conversion.output_data[c])

        plotter.__plot_base__(conversion, **kwargs)
        plotter.__plot_coast__(conversion, **kwargs)
        plotter.__plot_coverage__(conversion, *cpoints, **kwargs)
        plotter.__plot_vector__(conversion, *vpoints, **kwargs)


    def __plot_base__(self, conversion, frame=True, projection=None, region=None, **kwargs):
        
        _BASEMAP_ARGS = ['D','F','L','Td', 'Tm', 'U']
        _filter_base_kwargs =  { x:y for x,y in kwargs.items() if x in _BASEMAP_ARGS}
        for _arg in conversion.gmt_meta:
            if _arg in _BASEMAP_ARGS and _arg not in _filter_base_kwargs:
                _filter_base_kwargs[_arg] = conversion.gmt_meta[_arg]

        
        if region is None and conversion.gmt_region is None:
            region = region_bounds(conversion.region)
        elif region is None and conversion.gmt_region is not None:
            region = conversion.gmt_region
        elif region is None:
            region = DEFAULT_REGION


        if projection is None and conversion.gmt_projection is None:
            projection = DEFAULT_PROJECTION
        elif projection is None:
            projection = conversion.gmt_projection

        frame = frame or conversion.gmt_frame # differentiate from None
                   
        self.figure.basemap( region=region, projection=projection, frame=frame, **_filter_base_kwargs)
    def __plot_coast__(self, conversion,  **kwargs):

        _COAST_ARGS = ['A','C', 'D', 'G','I', 'N', 'S', 'W']
        _COAST_ARGS_ALT = ['area_thresh', 'lakes','resolution', 'land', 'rivers', 'borders',  'water', 'shorelines']
        _COAST_DICT = dict(zip(_COAST_ARGS_ALT, _COAST_ARGS))
        _filter_base_kwargs =  { x:y for x,y in kwargs.items() if x in _COAST_ARGS}
        for _arg in conversion.gmt_meta:
            if _arg in _COAST_DICT and _COAST_DICT[_arg] not in _filter_base_kwargs:
                _filter_base_kwargs[ _COAST_DICT[_arg] ] = conversion.gmt_meta[_arg]


        print (_filter_base_kwargs)
        self.figure.coast(**_filter_base_kwargs)
        
                
    def __plot_coverage__(self, conversion, *cpoints, **kwargs):
        def _filter_plot_pars(**pars):
            

        
        for cpoint in cpoints:
            if hasattr(cpoints,'gmt_meta')
        
        pass

    def __plot_vector__(self, conversion, *args, **kwargs):
        pass


            
            
            
            
            
        

    
        
    

    
        





def dmstodec(x):
    """ Utility Function to convert CDDMMSS.sssss
    to floating point decimal degrees
    
    where C is the Cardinal Direction (N/W/S/E) indicator for sign
    """

    c = 1.0 if x[0] in "NW" else -1.0
    d = float(x[1:3])
    m = float(x[3:5])
    s = float(x[5:])

    return c*(d+m/60.0 + s/3600.0)








def output_filename(output_type='v',
                    region='conus',
                    old_datum='ussd',
                    new_datum='nad27',
                    grid_spacing='900',
                    vdir='lon',
                    vclass='a',
                    vout='cd',
                    vunit='m',
                    surface=False, no_errors=True,*args, **kwargs):


    output_type = output_type.lower()
    if output_type not in ['v', 'c']:
        if output_type in 'vector':
            output_type = 'v'
        elif output_type in 'coverage':
            output_type = 'c'
        elif output_type in 'surface':
            output_type = 'v'
            surface = True
        else:
            raise ValueError("Output_type must be (v)ector, (c)overage, or (s)urface")

    prefix = ""
    if (output_type == 'v' and surface):
        prefix = "s"
    elif output_type == 'v':
        prefix = 'v'
    elif output_type == 'c':
        prefix = 'cv'
    
    if output_type == 'c' and surface:
        raise ValueError("No surface ready output type for coverage")

    
    vunit = vunit.lower()
    if vunit not in ['m','s']:
        if vunit in 'meters':
            vunit = 'm'
        elif vunit in 'arcseconds':
            vunit ='s'
        else:
            raise ValueError("Cannot Determine Intent of vunit = %s as (m)eters or arcsecond(s)"%vunit)
        
    vclass = vclass.lower()
    if vclass not in ['a','t','r','d']:
        if vclass in 'all':
            vclass = 'a'
        elif vclass in 'thinned':
            vclass = 't'
        elif vclass in 'dropped':
            vclass = 'd'
        elif vclass in 'rms':
            vclass = 'r'
        else:
            raise ValueError("Cannot Determine Intent of vclass = %s as (a)ll, (t)hinned, (d)ropped, or (r)ms"%vclass)
        
    vout = vout.lower()
    if vout not in ['gi','cd','dd']:
        if vout in 'gridinterp':
            vout = 'gi'
        elif vout in 'doubledifference':
            vout = 'dd'
        else:
            raise ValueError("Cannot determine intent of vout = %s as (g)rid(i)nterp or (d)ouble(d)ifference"%vout)
            
            
    vdir = vdir.lower()
    if vdir not in ['lon', 'lat', 'eht', 'hor']:
        if vdir in 'longitude':
            vdir = 'lon'
        if vdir in 'latitude':
            vdir = 'lat'
        if vdir in 'horizontal':
            if output_type == 'c':
                raise ValueError("No (hor)izontal coverage for output_type = (c)overage")
            vdir = 'hor'
        if vdir in 'height':
            vdir = 'eht'
        else:
            raise ValueError("Cannot determine intent of vdir = %s as (lon)gitude, (lat)itutde, (hor)izontal, or height (eht)"%vdir)

    if vunit == 's'  and vdir == 'eht':
        raise ValueError("No Height Difference for vdir of arcseconds")

    if not(no_errors):
        if output_type == 'c' and vdir == 'hor':
            raise ValueError("No Horizontal Coverage File")
        
        if output_type == 'c' and  vclass == 'r' and vdir == 'eht':
            raise ValueError("No Horizontal Difference for vout of rms")
        
        if ( (vout == 'dd') ^ (vclass == 'r') ) and output_type =='c':
            raise ValueError("vout=dd requires vclass=rms for coverage")
        
        if  vclass == 'r' and not vout == 'dd' and output_type == 'v':
            raise ValueError("vclass=rms requires vout=dd for vector")
        
        if output_type == 'c' and vout == 'gi':
            raise ValueError("No (g)rid(i)nterp vout for output_type = (c)overage")                
    

    if output_type == 'c':
        if  vclass in ['r','d','t']:
            ffile = "{prefix}{vclass}{vout}{vdir}.{old_datum}.{new_datum}.{region}.{grid_spacing}"
        else:
            ffile = "{prefix}{vclass}{vout}{vdir}.{old_datum}.{new_datum}.{region}"
    else:
        if surface or  vout in ['gi', 'dd'] or vclass in ['r','d','t']:
            ffile = "{prefix}{vunit}{vclass}{vout}{vdir}.{old_datum}.{new_datum}.{region}.{grid_spacing}"
        else:
            ffile = "{prefix}{vunit}{vclass}{vout}{vdir}.{old_datum}.{new_datum}.{region}"


    return ffile.format(prefix=prefix, **{'old_datum':old_datum,
                  'new_datum':new_datum,
                  'region':region,
                  'grid_spacing':grid_spacing,
                  'vunit':vunit,
                  'vclass':vclass,
                  'vout':vout,
                  'vdir':vdir})
    





#! /usr/bin/env python
from setuptools import setup
from numpy.distutils.core import Extension
from distutils.command.build   import build as DistutilsBuild
from distutils.command.install import install as DistutilsInstall
from os import listdir, path, environ
from datetime import datetime
import subprocess


from numpy.distutils.core import Extension, setup

# Development version is datetime + optional git hash
VERSION = environ.get('NC5NG_VERSION')
LOG_LEVEL = int(environ.get('NC5NG_DEBUG',3))
LOG_DEBUG = 5
LOG_INFO = 4
LOG_WARNING = 3
LOG_ERROR = 2
LOG_FAIL = 1
LOG_ALL = 0

def log (level, *msg):
    if level < LOG_LEVEL:
        print(*msg)

if not VERSION:
    VERSION="DEV"
    """
    now = datetime.now()
    VERSION="%s%s%s+%s%s"%(now.year, now.month, now.day, now.hour, now.minute)

    root_dir = path.abspath(path.dirname(__file__))
    if path.exists(path.join(root_dir, '.git')):
        cmd = ['git', 'rev-parse', '--verify', '--short', 'HEAD']
        git_hash = subprocess.check_output(cmd).decode().strip()
        VERSION = "%s.%s"%(git_hash, VERSION)

    """

    
               


PKG_INFO= {
    'version':VERSION,
    'description': "Python Package and Wrapper for the NADCON5 Datum Transformation Data and Utilities",
    'long_description': """Used to Generate and analyze Transformations of  US National Geodetic Survey Datums USSD, NAD27, and NAD83, and various other realizations. 

For More Information See: https://nc5ng.org

For Documentation See: https://docs.nc5ng.org/latest
""",
    'author':"Andrey Shmakov",
    'author_email':"akshmakov@gmail.com",
    'url':"https://nc5ng.org",
    'download_url':"https://github.com/nc5ng/nadcon5-ng",
}

## Selectively disable modules from processing
## List Fortran Files not intended for Build Here
DISABLED_MODULES = [ 
    ]

## Default Options for F2PY Invocation
F2PY_DEFAULT_OPTS = { "include_dirs" : ["src/Subs"] }

## Module Specific Includes
MODULE_INCLUDES = {
    'nc5ng.core.bwplotvc': ['src/Subs/plotcoast.f'],
    'nc5ng.core.bwplotcv': ['src/Subs/plotcoast.f'],
    'nc5ng.core.coplot': ['src/Subs/plotcoast.f'],
    'nc5ng.core.coplotwcv': ['src/Subs/plotcoast.f'],
    'nc5ng.core.gridstats' : ['src/Subs/select2_dbl.for']
    }

## Module Specific Options
MODULE_OPTS = {
    'nc5ng.core.bwplotvc': {}, # { 'library_dirs' : ['build/lib'],
                             #'libraries': ['plotcoast' ]},
    'nc5ng.core.bwplotcv':{},
    'nc5ng.core.coplot': {},
    'nc5ng.core.coplotwcv': {},
}


class MakeBuilder(DistutilsBuild):
    def run(self):
        DistutilsBuild.run(self)

def _merge_dict(x,y):
    z = x.copy()
    z.update(y)
    return z

##
## Make The KWARGS aka a Dictionary
## suitable for passing to numpy.distutils.core.Extension
##
def mk_fortran_extension_kwargs(src_file, pkg, sig_dir = None):
    """ Create kwargs for the given fortran extension
    \param src_file the name of the source file to be built
    \param pkg the root package that will be built
    \param sig_dir directory containing signature file
    \return dictionary containing KWARGS
    """
        
    root, ext = path.splitext(src_file)
    src_dir, name = path.split(root)

    mod_name = ("%s.%s"%(pkg, name)).strip()

    if ext not in ['.f', '.for']:
        log (LOG_WARNING,src_file, " is not a valid fortran file, ignoring ")
        return None

    if mod_name in DISABLED_MODULES:
        log (LOG_WARNING, mod_name, " is explicitly disabled in setup.py , ignoring ")
        return None

    sig_file_name = name + ".pyf"
    sig_file = None

    f2py_opts = MODULE_OPTS.get( mod_name, F2PY_DEFAULT_OPTS )

    log (LOG_INFO, "Found Valid src_file ", src_file, " for package ", pkg,
           " , with sig_dir ", sig_dir, " module name ", mod_name, " f2py_opts ", f2py_opts)

    sources = [ src_file ]
    
    if sig_dir and path.exists(path.join(sig_dir, sig_file_name)):
        sig_file = path.join(sig_dir, sig_file_name)
        print ("Found signature file ", sig_file)
    elif path.exists(path.join(src_dir, sig_file_name)):
        sig_file = path.join(src_dir, sig_file_name)
        print ("Found signature file ", sig_file)

    if sig_file is not None:
        sources = [ sig_file, src_file ]

    sources = MODULE_INCLUDES.get( mod_name , []) + sources

    return _merge_dict({'name': mod_name,
                        'sources' : sources}
                       , f2py_opts)


## Setup Constants for our packages
ROOT_PKG = 'nc5ng'
    
CORE_PKG = 'nc5ng.core'
CORE_SRC_DIR = 'src/Subs'
CORE_PKG_DIR = 'nc5ng/nc5core'
CORE_PROGRAMS = [mk_fortran_extension_kwargs(path.join(CORE_SRC_DIR, f),
                                             CORE_PKG,
                                             CORE_PKG_DIR)
                 for f in listdir(CORE_SRC_DIR)]


DATA_PKG = 'nc5ng.nc5data'
DATA_PKG_DIR = 'nc5ng/nc5data'


BUILD_PKG = 'nc5ng.nc5build'
BUILD_PKG_DIR = 'nc5ng/nc5build'


PACKAGES = [ROOT_PKG, CORE_PKG, DATA_PKG]#, BUILD_PKG]


## Run Setup 
if __name__ == '__main__':
    fortran_extensions = []

    ## Run Through All Extensions
    for kwargs in CORE_PROGRAMS:
        if kwargs is not None:
            fortran_extensions.append(Extension(**kwargs))
    
    setup(name = 'nc5ng-core',
          packages = PACKAGES,
          ext_modules = fortran_extensions,
          **PKG_INFO
          )
          
    #setup(**configuration(top_path='').todict())



from os.path import basename
import logging




class DataPointType(type):
    """ Metaclass for DataPoints, defines class creation and class/hierarchy member variables 
    
    The meta-class in part, hides some of the more rote requirements of our datapoint type
    from the actual object hierarchy

    This allows some magic like run-time casting to the correct datapoint without knowing the 
    type specifically, and a persisitent library-wide memory backed database for quick retrieval and to minimize replication

    To understand the meta class, there are a number of tutorials available online, roughly speaking this class is what is used to "create" the DataPoint class, and allows us to manipulate the class without needing any implementation details. 


    Defining a new DataPointType Hierarchy simply requires using this class as the metaclass

        class NewDataPoint(metaclass=DataPointType):
          pass

    By creating a new DataPointType, the following changes will be done to the final class

      - The type will have a database (dictionary) of types registered with the base class
      - Each type and subtype will have a point container (default set) created
      - The shorthand name will be generated from the class name
      - Instance Creation will be overidden and allow creation of any other data type by specifting the type shorthand as an argument
      

    Class Configuration:

    Each new type has some meta-configuration available

    1. To override data point registration 
       - Create a `@classmethod`  `__register__(cls, point)` to overide how a new point is registered/saved
       - Create a class member `_point_store` to change the underlying storage type (from set)
    2. Override shorthand name by specific '_type_shorthand' explicitly in the class
    
    
    Any class that uses this type as a metaclass will be registered 
    """
    @classmethod
    def __prepare__(metacls, name, bases, **kargs):
        """ Prepare the new class, here for completeness
        """
        logging.debug("Preparing Class %s"%name)
        return super().__prepare__(name, bases, **kargs)


    @property
    def type_shorthand(cls):
        """ Get the class shorthand name"""
        return cls._type_shorthand
    
    @property
    def point_store(cls):
        """ Return the type-specific Point Buffer"""
        return cls._point_store

    @property
    def point_database(cls):
        """ Return the Root Database of all DataPoints in this Hierarchy"""
        return cls._cbdb
    
    
    def __new__(metacls, name, bases, namespace,  **kargs):
        """ Create a new data point type, called on class load

        Creates class attributes level point set for storage
        
        metaclass __new__ is executed on load time for every
        class that uses it, it is executed after __prepare__ which
        constructs the class object
        """

    
        logging.debug("Creating Class %s"%name)
        cls = super().__new__(metacls, name, bases, namespace)
        
        if not(hasattr(cls, '_cbdb')):
            logging.debug("Creating Data Point Database")
            cls._cbdb = dict()


        # Shorthand Name
        cls._type_shorthand =  name.lower()
        while cls._type_shorthand in cls._cbdb:
            cls._type_shorthand = cls._type_shorthand + "_" # in case of name conflict, add underscore

        # Point Storage
        cls._point_store = namespace.get('_point_store', set() )
        
        
        logging.debug("Registering new Data Point Type %s with shorthand %s"%(name, cls._type_shorthand))

        
        cls._cbdb[cls.type_shorthand]={'type':cls, 'points': cls._point_store }
        return cls

    
    def __init__(cls, name, bases, namespace):
        """ Initialize a new FileBacked Class

        This is a slot method for class creation, __init__ is called when class is defined (load time)

        \param cls - reference to new type, similiar to @classmethod
        \param name - new class name
        \param bases - base classes
        \param namespace - new class attributes
        \param Parser - BaseFileParser underlying this type
        \param **kwargs - keywords passed to Parser initialization
        
        """        
        logging.debug("Creating Data Point Class %s"%name)

        super().__init__(name, bases, namespace)


    def __register__(cls, point):
        """ Register a point with the class buffer """
        cls._point_store.add(point)

    def __call__(cls, *args,  **kw):
        """ Create a new Point in this hierarchy"""

        """
        if typename is not None:
            if typename not in cls._cbdb:
                raise TypeError("Invalid Data Point Type with Shorthand %s"%typename)
            cls = cls._cbdb[typename]['type']
        """
        if args and (args[0] in cls.point_database.keys()):
            typename=args[0]
        elif kw and ('type' in kw):
            typename = kw['type']
        else:
            typename = cls.type_shorthand


        if typename == cls.type_shorthand:
            point = super().__call__(*args, **kw)
            if not getattr(point, 'ephemeral', False):
                cls.__register__(point)
        elif typename in cls.point_database.keys():
            point = cls.point_database[typename]['type'].__call__(*args, **kw)
        else:
            return None

        return point

    
    
class DataPoint(metaclass=DataPointType):
    """ 
    Base Data Point Type
    """
    def __init__(self, *args, **kwargs):
        self._data = {}

        # for copy
        if args:
            if issubclass(args[0].__class__, DataPoint):
                kwargs = args[0]._data
        #initialize
        for k,v in kwargs.items():
            self._data[k] = v

    
    def __getattr__(self, name):
        if name in self._data:
            return self._data[name]
        else:
            raise AttributeError()
    

    @property
    def data(self):
        return self._data


    def __getitem__(self,name):
        if name in self._data:
            return self._data[name]
        else:
            raise KeyError()
    """
    def __eq__(self, other):
        if not issubclass(other.__class__, DataPoint) or self.type_shorthand != other.type_shorthand:
            return False
        else:
            for k in self.data:
                if not k in other.data or self.data[k] != other.data[k]:
                    return False
            for k in other.data:
                if not k in self.data:
                    return False
        return True
    """

    def __ge__(self, other):
        """ Superset """
        if  not issubclass(other.__class__, DataPoint):
            return False
        return other.__le__(self)
        
        
    def __gt__(self, other):
        """ Strict Superset """
        if not issubclass(other.__class__, DataPoint):
            return False
        return other.__lt__(self)
        
    def __le__(self, other):
        """ Subset, each element in our data is contained in the other
        """
        if not issubclass(other.__class__, DataPoint):
            return False

        if len(self.data) > len(other.data):
            return False
        else:
            for k in self.data:
                if not k in other.data or self[k] != other[k]:
                    return False
        return True

    def __lt__(self, other):
        """ Strict Subset 
        """
        if not issubclass(other.__class__, DataPoint):
            return False
        
        if len(self.data) >= len(other.data):
            return False

        ## equality requires same length, already checked equality 
        return self.__le__(other)

    
    def __contains__(self, other):
        if not issubclass(other.__class__, DataPoint):
            try:
                k,v = other
                if not k in self.data:
                    return False                
                return self[k] == v
            except (TypeError, ValueError):
                return other in self.data
        else:
            for k in other.data:
                if not k in self.data:
                    return False
        return True

    def __repr__(self):
        return "DataPoint('" + self.__class__.type_shorthand + "', **"+repr(self._data)+")"

    def __str__(self):
        return "Data Point('" + self.__class__.type_shorthand + "') : " + str(self._data)
    
    def __bool__(self):
        return self.data.__bool__()
    

    
class MetaMixin(object):
    """
    Mixin base type for standard meta information
    """    
    
    def _safe_meta_property(index):
        def getter(self):
            if self.meta and index in self.meta:
                return self.meta[index]
            else:
                return None
        return property(getter)
    
    
    
    @property
    def meta(self):
        if hasattr(self, '_meta'):
            return self._meta
        elif hasattr(self, '_data'):
            if 'meta' not in self._data:
                self._data['meta'] = {}
            return self._data['meta']
        else:
            return None

    source = _safe_meta_property('source')
    old_datum = _safe_meta_property('old_datum')
    new_datum = _safe_meta_property('new_datum')
    region = _safe_meta_property('region')
    subregion = _safe_meta_property('subregion')

    @property
    def shorthand(self):
        if self.source:
            return basename(self.source).split('.')[0]




class DataContainerMixin(object):
    """
    Mixin type for containers of DataPoints

    """
    
    @property
    def data(self):
        """
        Get underlying raw data 
        """
        if hasattr(self, '_data'):
            if 'data' not in self._data:
                self._data['data'] = None
            return self._data['data']
        else:
            return None

    @property
    def points(self):
        """
        Get indexed data points
        """
        if hasattr(self, '_indexed_data'):
            return self._indexed_data.values()
        else:
            return None
    
    @property
    def indices(self):
        """
        index values
        """
        if hasattr(self, '_indexed_data'):
            return self._indexed_data.keys()
        else:
            return None

    @property
    def index_tag(self):
        """
        DataPoint field used for indexing
        """
        if hasattr(self, '_index_tag'):
            return self._index_tag
        else:
            return None
    

    def __iter__(self):
        """
        Default Iterator
        """
        for index in self.indices:
            yield self[index]
            
    def __getitem__(self, index):
        if index in self.indices:
            return self._indexed_data[index]
        else:
            return None
    def __init__(self, *args, **kwargs):
        if hasattr(self.__class__, 'parser'):
            self._data = self.__class__.parser.fromfile(*args, **kwargs)
            
        self.__init_indexed_data__()

    def __len__(self):
        return len(self.data)

    def __contains__(self, item):
        if item in self.indices:
            return True
        elif item in self.indexed_data:
            return True
        return False
        






class GMTMixin(object):

        
    @property
    def plot_args(self):
        return self.__mk_plot_args__()
    
    def plot(self, figure, **kwargs):
        pkwargs = self.plot_args
        pkwargs.update(kwargs)
        figure.plot(**pkwargs)

    
        


        

class GMTPointMixin(object):

    @property
    def gmt_data(self):
        return self.__mk_gmt_data__()

    
    
    




class GMTMetaMixin(object):
    def _safe_gmt_meta_property(index):
        def getter(self):
            if self.gmt_meta and index in self.gmt_meta:
                return self.gmt_meta[index]
            else:
                return None
        return property(getter)

    @property
    def gmt_meta(self):
        if not(hasattr(self, '_gmt_meta')):
            self._gmt_meta = {}
        return self._gmt_meta



    gmt_region = _safe_gmt_meta_property('region')
    gmt_projection = _safe_gmt_meta_property('projection')
    gmt_rivers= _safe_gmt_meta_property('rivers')
    gmt_borders= _safe_gmt_meta_property('borders')
    gmt_water= _safe_gmt_meta_property('water')
    gmt_shorelines = _safe_gmt_meta_property('shorelines')
    gmt_land = _safe_gmt_meta_property('land')
    gmt_resolution = _safe_gmt_meta_property('resolution')
    gmt_frame = _safe_gmt_meta_property('frame')
